# Author: Alejandro Eduardo Cruz Paz
# GraphSnit 1.0 for the tcllib package 
# 

package require Tcl 8.5
package require snit 

package require struct::prioqueue
package require struct::stack
package require struct::set

source graph.tcl
source disjointset.tcl

package require struct::disjointset
package require struct::graph 2.3

snit::type graph {

  component structure

  #Constructor for this abstract data type
  constructor {args} {
      if {$args eq {}} {
	  set structure [::struct::graph "$self\_tcl"]
      } else {
	  # handling graph copy
	  if {[lsearch -regexp $args {=|:=|as}] != -1} {
	      set sourceg [lindex $args end]
	      set args [lreplace $args end end "$sourceg\_tcl"]
          } 
	  set structure [::struct::graph "$self\_tcl" $args]
      }
  }

  #Destructor for this abstract data type 
  destructor {
    catch {
      $structure destroy
      $self\_tcl destroy
    }
  }
  
  #Returns the Snit namespace this instance is asociated with
  method instancedata {option} {
      switch $option {
	  -namespace { return $selfns }
	  -snithandler { return $self }
	  -tclhandler { return $self\_tcl }
	  default { return -code error "Bad option for method instance data \
valid options are: -namespace, -snithandler, and -tclhandler"}
      }
  }

  method VerifyWeightsAreOK {} {
      set arcs [$self arcs]
      foreach arc $arcs {
	  catch {[$self arc getweight $arc]} msg
	  if {$msg eq "Arc $arc has not been assigned a weight yet"} { 
	      return 0
	  }
      }
      return 1
  }

  # Dijkstra algorithm implementation for finding the shortest path 
  # between two nodes
  # Input
  #  |node flag1 flag2|
  # where flag1 is -arcmode and it's options are undirected or directed, 
  # and flag2 is -outputformat and it's options are tree or distances.
  # Output
  #  two kinds of lists, a set of nodes constructing a minimum spanning tree, 
  #  pairs of nodes-values regarding the final value for each node  
  # UNDIRECTED/DIRECTED MODES
  method dijkstra { node args} {
    # verify arguments
    if {[llength $args] > 4} {
      return -code error "Wrong number of arguments. Should be -arcmode and -outputformat together with a related option"
    }
    # verify this graph is weighted
    if {![$self isweighted?]} {
	return -code error "This graph is not weighted. This algorithm requires the graph to be weighted"
    }
    # verify there are no arcs without a weight assigned
    if {![$self VerifyWeightsAreOK]} {
        return -code error "There are some arcs on this graph with no assigned weight"
    }

    set option1index [lsearch $args -arcmode]
    set option2index [lsearch $args -outputformat]
    if {$option1index == -1} {
      #assume undirected as the default arc mode for the algorithm to work with
      set option1 undirected
    } else {
      set option1t [lindex $args [incr option1index]]
      if {$option1t eq {undirected} || $option1t eq {directed}} {
        set option1 $option1t
      } else {
        return -code error "Bad option for -arcmode, options are undirected|directed"
      }
    }

    if {$option2index == -1} {
      #assume tree as the default return format in 
      #absence of a explicitly declared return format
      set option2 tree
    } else {
      set option2t [lindex $args [incr option2index]]
      if {$option2t eq {tree} || $option2t eq {distances}} {
        set option2 $option2t
      } else {
        return -code error "Bad option for -outputformat, options are tree|distances"
      }
    }


    # strtnode flag set to zero
    set strtnode 0
    ::struct::prioqueue -dictionary Q
    foreach currentnode [$self nodes] {

      # initialize the distance and previous atributes of the nodes on this graph
      if {$currentnode eq $node} {
        # on the node from which the algorithm is going to be executed
        set distance($node) 0
        set previous($node) {none}
        # start node exists on graph, then strtnode set to one
        set strtnode 1
        Q put $currentnode $distance($currentnode)
      } else {
        # on the other nodes
        set distance($currentnode) [expr {(2 << 30)-1}]
        set previous($currentnode) {undefined}
      }
    }
    if {!$strtnode} {
      Q destroy
      error "The node provided as the initial node does not exists on this graph"
    }
  
    while {[Q size] > 0} {
      set u [Q get]
      set extracted($u) 1
      # two different behaviours depending on the option2 selection
      switch $option1 {
       {undirected} { set arcsreach [$self arcs -adj $u] }
       {directed} { set arcsreach [$self arcs -out $u] }
      }
       # work with every neighbor of the current node 
      foreach {arc} $arcsreach {
        set arcost [$self arc getweight $arc]
        set v [$self arc target $arc]
        #if option2 is -undirected
        if {$option1 eq {undirected}} {
          # a cycle or the arc is pointing in the direction of the node 'u'
          if {$u eq $v} {
            set v [$self arc source $arc]         
          }
        }
        set res [expr $arcost + $distance($u)]
        # adjust the weights in the neighbors of the current node
        if {$res < $distance($v)} {
          set distance($v) $res 
          set previous($v) $u 
          if {![info exists extracted($v)]} {
            Q put $v $distance($v)       
          }
        } 
      }
    }

    switch -exact $option2 {
      {tree}  {     
            # bottom-up construction of the path to the origin node
            foreach cnode [$self nodes] {
               set u $cnode
               while {1} {
		 if {$u eq "undefined"} {
		     break
		 }
                 if {$u eq $node} {
                     lappend listofprevious($cnode) $u
                     break
                    }
                 if {![string match $u $cnode]} {
                     lappend listofprevious($cnode) $u
                    }
                 set u $previous($u)
               }
            }  
            set results [array get listofprevious]
            }
      {distances} { set results [array get distance]}
    }
    Q destroy

    return $results
  }

  # returns the distance between any two nodes in the graph, and such a 
  # distance is selected
  # on basis on the shortest path (weights sum) between those two nodes
  # Input
  # Two, origin node and the destination node
  # Output
  # returns the distance between these two nodes; origin node and the 
  # destination node.
  method distance { args } {
     if { [llength $args] != 2 } {
        error "Wrong number of arguments, need to be exactly two. The origin node and the destination node"
     } 
     set firstnode [lindex $args 0]
     set secondnode [lindex $args 1]
     #the distance from a node to itself is zero
     if {$firstnode eq $secondnode} {
       return 0
     }
     set resultinglist [$self dijkstra $firstnode tree]
     foreach {node list} $resultinglist {
        if {$node eq $secondnode} {
          return [llength $list]
        }
     }
  }

  #find the eccentricity of a node
  #the eccentricity of a node is the greatest distance between v and any other vertex.
  method eccentricity {args} {
      if {[llength $args] == 0} {
	  return -code error "Wrong # of arguments. A node at least must be provided"
      }
      if {[llength $args] > 3} {
	  return -code error "Wrong # of arguments. No more than three arguments are allowed"
      }
    set node [lindex $args 0]
    set index [lsearch $args {-arcmode}]
    #check arguments for type of graph considerations
    if {$index == -1} {
      set resultinglist [$self dijkstra $node -outputformat \
                        tree -arcmode undirected]
    } else {
      set graphtype [lindex $args [incr index]] 
      set resultinglist [$self dijkstra $node -outputformat \
                        tree -arcmode $graphtype]
    }
    set ecc 0
    foreach {nodec list} $resultinglist {
       if {$ecc < [llength $list]} {
          set ecc [llength $list]
       } 
    }
    return $ecc  
  }

  #return the radius of this graph
  method radius { {args {}} } {
    # call GenericaRadiusDiameter passing a code fragment for replacement, 
    # as diameter and radius are very similar
    return [$self GenericRadiusDiameter {expr $ecctmp < $ecc} [expr {(2 << 30)-1}] $args]
  }

  #return the diameter of this graph
  method diameter { {args {}} } {
    # call GenericaRadiusDiameter passing a code fragment for replacement, 
    # as diameter and radius are very similar
    return [$self GenericRadiusDiameter {expr $ecctmp > $ecc} 0 $args]
  }

  #helper for diameter and radius method
  method GenericRadiusDiameter {fragmentedcode number args} {
    set ecc $number
    foreach node [$self nodes] {
      set ecctmp [$self eccentricity $node $args] 
      if {[eval $fragmentedcode]} {
        set ecc $ecctmp
      }
    }
    return $ecc
  }

  # Method prim. Prim's algorithm implementation for finding minimum spanning trees on a graph
  # Input
  #   |nodesource| is the node from which the algorithm will be executed
  # Output
  #   A list of nodes which conform the resulting minimum spanning tree
  # UNDIRECTED MODE
  method prim {nodesource} {
 
    # verify there are no arcs without a weight assigned
    if {![$self VerifyWeightsAreOK]} {
        return -code error "There are some arcs on this graph with no assigned weight"
    }   
    #Construct a priority queue
    struct::prioqueue -dictionary Q
    set nodeset [$self nodes]
    set nodetest 0
    foreach {node} $nodeset {
      if {$nodesource eq $node} {
        set nodetest 1
      }
    }
    
    if {$nodetest == 0} {
      Q destroy
      error "The node you provided as the initial node does not exists on this graph"     
    }  
    Q put $nodesource 0 
    #while the queue doesn't run out of nodes
    while {[Q size] > 0} {
      set arc [Q peekpriority 1]
      set v [Q get 1]
      #check if the node is already on the tree
      if {[info exist resultingtree($v)]} {
        set currentv $resultingtree($v)
        #check the assigned value to node v, if a lesser value is found, change it
        if {$arc < $currentv} {
          set resultingtree($v) $arc
        }
      } else {
        set resultingtree($v) $arc
        #considering nodes as nondirected nodes
        set arcsreach [$self arcs -adj $v]
        #put each neighbour with it's value onto the priority queue
        foreach {arc} $arcsreach {
          set arcost [$self arc getweight $arc]
          if {[$self arc target $arc] eq {}} {
            set node [$self arc source $arc]
          } else {
           set node [$self arc target $arc]
          }
          Q put $node $arcost
        }
      }
    } 
    Q destroy
    return [array get resultingtree]
  }

  # Method kruskal. Another algorithm for finding minimum spanning trees
  method kruskal { } {
    # priority queue for saving the arcs together with it's associated weights
   ::struct::prioqueue -dictionary Q
   ::struct::disjointset dset
   set forest {}   

   foreach {node} [$self nodes] {
      dset makeset $node
   }

   foreach {edge cost} [$self arc weights] {
      Q put $edge $cost
   }

   while {[Q size] > 0} {
      set e [Q get]
      set nsource [$self arc source $e]
      set ntarget [$self arc target $e]
      set set1 [dset find $nsource -sets]
      set set2 [dset find $ntarget -sets]
      
      if {![dset equals $set1 $set2]} {
        set forest [concat $forest "$nsource $e $ntarget"]
        dset findunitereplace $nsource $ntarget
      }
    }
    Q destroy
    dset destroy
    return $forest

  }

  # Method tarjan. Tarjan algorithm is an algorithm for finding the strongly connected components
  # of a digraph
  # In a strongly connected component, for every two nodes U,V there is always a path from U to V
  # and from V to U.
  # If the resulting set of connected components is empty, then the digraph is a directed acyclic
  # graph
  # Input:
  #   |startnode| is the vertex from which the algorithm will be executed
  # Output:
  #   A list containing the strongly connected components
  # DIRECTED MODE

  method tarjan {startnode} {
    # build an stack object
    ::struct::stack S
    # result saves the strongly connected components found in each recursive step if method tarjansub
    variable result    
    set result {}
    variable index
    # calling method tarjansub which will do all the job, starting on startnode
    $self tarjansub $startnode 0
    unset index
    # destroy stack object
    S destroy
    return $result
  }

#
#proc foo_rec {depth} {upvar bar bar; set bar($depth) x; if {$depth < #100} {foo_rec [expr {$depth+1}]}}
#
  method tarjansub {startnode counter} {
    variable index
    variable lowlink
    variable inS
    variable result
    set component {}

    set index($startnode) $counter
    set lowlink($startnode) $counter
    incr counter
    S push $startnode
    # track nodes added to stack
    set inS($startnode) 1
    set E [$self arcs -out $startnode]

    foreach {arc} $E {
       set tnode [$self arc target $arc]
       if {![info exists index($tnode)]} {
          # DFS from node tnode, from startnode at the first invocation of method tarjansub
          $self tarjansub $tnode $counter
          set lowlink($startnode) [$self Min $lowlink($startnode) $lowlink($tnode)]
       } elseif { [info exists inS($tnode)] } {
          set lowlink($startnode) [$self Min $lowlink($startnode) $index($tnode)]
       }
    }

    # if startnode on this recursion level is root of an strongly connected component
    if {$lowlink($startnode) == $index($startnode)} {
      while {1} {
        set v [S pop]
        lappend component $v
        if {$v eq $startnode} {
          break
        }
      }
      lappend result $component
    }
  }

  #Method for tracking the visited nodes. Helper of method connectedness
  proc savingproc {a g n} { global t ; lappend t $n ; return}
  
  #This method applies the DFS algorithm for searching the
  # number of connected components this graph has.
  method connectedness {{restofnodeset {}} {partition {}}} {
    global t
    if {$restofnodeset == {}} {
      set nodeset [$self nodes]
    } else {
      set nodeset $restofnodeset
    }
    set firstnode [lindex $nodeset 0]
    $self walk $firstnode -type dfs -dir forward -command savingproc
    set direction1 $t ; set t {}
    $self walk $firstnode -type dfs -dir backward -command savingproc
    set traversed [::struct::set union $direction1 $t]; set t {}
    set difference [::struct::set difference $nodeset $traversed]
    lappend partition $traversed
    if {$difference eq {}} { return [lindex $partition 0] }
    foreach node $traversed {
      set index [lsearch $nodeset $node]
      set nodeset [lreplace $nodeset $index $index]
    }
    $self connectedness $difference $partition
  }

  # This method checks if after removing the node which is passed as argument
  # the number of connected components of the graph augments, in which case
  # means the node provided is indeed a cutvertex of the graph.
  method iscutvertex? {pseudocutvertex} {
    set nodeset [$self nodes]
    if {[lsearch $nodeset $pseudocutvertex] == -1} {
      return -code error "The node $pseudocutvertex doesn't exists on this graph"
    }
    set graphcopy [$self CopyGraph "copy"]
    set ctedcomp_prev [$graphcopy connectedness]
    $graphcopy node delete $pseudocutvertex
    set ctedcomp_actu [$graphcopy connectedness]
    $graphcopy destroy
    if {$ctedcomp_prev > $ctedcomp_actu} {
      return 1
    } else {
      return 0
    }
  }

  # This method checks if after removing the edge which is passed as argument
  # the number of connected components of the graph augments, in which case
  # means the edge provided is a bridge of the graph.
  method isbridge? {pseudobridge} {
    set edgeset [$self arcs]
    if {[lsearch $edgeset $pseudobridge] == -1} {
      return -error code "The edge $pseudobridge doesn't exists on this graph"
    }
    set graphcopy [$self CopyGraph "copy"]
    set ctedcomp_prev [$graphcopy connectedness]
    $graphcopy arc delete $pseudobridge
    set ctedcomp_actu [$graphcopy connectedness]
    $graphcopy destroy
    if {$ctedcomp_prev > $ctedcomp_actu} {
      return 1
    } else {
      return 0
    }
  }

  method Min { first second } {
    if {$first > $second} {
      return $second
    } else {
      return $first
    }
  }

  method CopyGraph {name} {
    if {[$self isweighted?]} {
      set option "-weighted"
    } else {
      set option "-unweighted"
    }    
    namespace eval :: graph $name $option = $self { }
  }

  delegate method nodes to structure
  delegate method arcs to structure
  delegate method node to structure
  delegate method arc to structure
  delegate method append to structure
  delegate method deserialize to structure
  delegate method lappend to structure
  delegate method get to structure
  delegate method getall to structure
  delegate method keys to structure
  delegate method keyexists to structure
  delegate method serialize to structure
  delegate method set to structure
  delegate method swap to structure
  delegate method unset to structure
  delegate method walk to structure
  delegate method destroy to structure
  delegate method isweighted? to structure
}
