#ifndef _TCL_RTLD
#define _TCL_RTLD

#include <sys/types.h>
#include <sys/queue.h>

#include "elf-hints.h"
#include <link.h>
#include <stddef.h>
#include <assert.h>

#include "rtld_machdep.h"
#include "defs.h"

#ifndef STANDARD_LIBRARY_PATH
#define STANDARD_LIBRARY_PATH	"/lib"
#endif

#define NEW(type)	((type *) xmalloc(sizeof(type)))
#define CNEW(type)	((type *) xcalloc(sizeof(type)))

/* We might as well do booleans like C++. */
typedef unsigned char bool;
#define false	0
#define true	1

extern size_t tls_last_offset;
extern size_t tls_last_size;
extern size_t tls_static_space;
extern int tls_dtv_generation;
extern int tls_max_index;

extern Elf_Addr _GLOBAL_OFFSET_TABLE_[];

struct stat;
struct Struct_Obj_Entry;

/* Lists of shared objects */
typedef struct Struct_Objlist_Entry {
    STAILQ_ENTRY(Struct_Objlist_Entry) link;
    struct Struct_Obj_Entry *obj;
} Objlist_Entry;

typedef STAILQ_HEAD(Struct_Objlist, Struct_Objlist_Entry) Objlist;

/* Types of init and fini functions */
typedef void (*InitFunc)(void);

/* Lists of shared object dependencies */
typedef struct Struct_Needed_Entry {
    struct Struct_Needed_Entry *next;
    struct Struct_Obj_Entry *obj;
    unsigned long name;		/* Offset of name in string table */
} Needed_Entry;

/* Lock object */
typedef struct Struct_LockInfo {
    void *context;		/* Client context for creating locks */
    void *thelock;		/* The one big lock */
    /* Debugging aids. */
    volatile int rcount;	/* Number of readers holding lock */
    volatile int wcount;	/* Number of writers holding lock */
    /* Methods */
    void *(*lock_create)(void *context);
    void (*rlock_acquire)(void *lock);
    void (*wlock_acquire)(void *lock);
    void (*rlock_release)(void *lock);
    void (*wlock_release)(void *lock);
    void (*lock_destroy)(void *lock);
    void (*context_destroy)(void *context);
} LockInfo;

/**
 * Shared object descriptor.
 *
 * Items marked with "(%)" are dynamically allocated, and must be freed
 * when the structure is destroyed.
 *
 * CAUTION: It appears that the JDK port peeks into these structures.
 * It looks at "next" and "mapbase" at least.  Don't add new members
 * near the front, until this can be straightened out.
 */
typedef struct Struct_Obj_Entry {
    /*
     * These two items have to be set right for compatibility with the
     * original ElfKit crt1.o.
     */
    Elf_Word magic;		///< Magic number (sanity check)
    Elf_Word version;		///< Version number of struct format


    char *path;			///< Pathname of underlying file (%)
    char *origin_path;		///< Directory path of origin file
    int refcount;
    int dl_refcount;		///< Number of times loaded by elf_dlopen

    /* These items are computed by Tcl_MapObjectFromFile() or by digest_phdr(). */
    caddr_t mapbase;		///< Base address of mapped region
    size_t mapsize;		///< Size of mapped region in bytes
    size_t textsize;		///< Size of text segment in bytes
    Elf_Addr vaddrbase;		///< Base address in shared object file
    caddr_t relocbase;		///< Relocation constant = mapbase - vaddrbase
    const Elf_Dyn *dynamic;	///< Dynamic section
    caddr_t entry;		///< Entry point
    const Elf_Phdr *phdr;	///< Program header if it is mapped, else NULL
    size_t phsize;		///< Size of program header in bytes
    const char *interp;		///< Pathname of the interpreter, if any

    /* TLS information */
    int tlsindex;		///< Index in DTV for this module
    void *tlsinit;		///< Base address of TLS init block
    size_t tlsinitsize;		///< Size of TLS init block for this module
    size_t tlssize;		///< Size of TLS block for this module
    size_t tlsoffset;		///< Offset of static TLS block for this module
    size_t tlsalign;		///< Alignment of static TLS block

    /* Items from the dynamic section. */
    Elf_Addr *pltgot;		///< PLT or GOT, depending on architecture
    const Elf_Rel *rel;		///< Relocation entries
    unsigned long relsize;	///< Size in bytes of relocation info
    const Elf_Rela *rela;	///< Relocation entries with addend
    unsigned long relasize;	///< Size in bytes of addend relocation info
    const Elf_Rel *pltrel;	///< PLT relocation entries
    unsigned long pltrelsize;	///< Size in bytes of PLT relocation info
    const Elf_Rela *pltrela;	///< PLT relocation entries with addend
    unsigned long pltrelasize;	///< Size in bytes of PLT addend reloc info
    const Elf_Sym *symtab;	///< Symbol table
    const char *strtab;		///< String table
    unsigned long strsize;	///< Size in bytes of string table

    const Elf_Addr *buckets;	///< Hash table buckets array
    unsigned long nbuckets;	///< Number of buckets
    const Elf_Addr *chains;	///< Hash table chain array
    unsigned long nchains;	///< Number of chains

    const char *rpath;		///< Search path specified in object
    Needed_Entry *needed;	///< Shared objects needed by this one (%)

    InitFunc init;		///< Initialization function to call
    InitFunc fini;		///< Termination function to call

    bool mainprog;		///< True if this is the main program
    bool rtld;			///< True if this is the dynamic linker
    bool textrel;		///< True if there are relocations to text seg
    bool symbolic;		///< True if generated with "-Bsymbolic"
    bool bind_now;		///< True if all relocations should be made first
    bool traced;		///< Already printed in ldd trace output
    bool jmpslots_done;		///< Already have relocated the jump slots
    bool init_done;		///< Already have added object to init list
    bool tls_done;		///< Already allocated offset for static TLS

    struct link_map linkmap;	///< for GDB and dlinfo()
    Objlist dldags;		///< Object belongs to these dlopened DAGs (%)
    Objlist dagmembers;		///< DAG has these members (%)
    dev_t dev;			///< Object's filesystem's device
    ino_t ino;			///< Object's inode number
    struct Struct_Obj_Entry *next;    ///< Pointer to the next member of global list of loaded objects
    struct Struct_Obj_Entry *prev;    ///< Pointer to the previous member of the global list of loaded objects
} Obj_Entry;

#define RTLD_MAGIC	0xd550b87a
#define RTLD_VERSION	1

/**
 * Symbol cache entry used during relocation to avoid multiple lookups
 * of the same symbol.
 */
typedef struct Struct_SymCache {
    const Elf_Sym *sym;		///< Symbol table entry
    const Obj_Entry *obj;	///< Shared object which defines it
} SymCache;

//TODO void		 _rtld_error(const char *, ...) __printflike(1, 2);
Obj_Entry	*Tcl_MapObjectFromFile(int, const char *, const struct stat *);
Obj_Entry       *Tcl_MapObjectFromMemory(const void *buffer, int nbytes);
void		*xcalloc(size_t);
void		*xmalloc(size_t);
void		*xrealloc(void *, size_t);
char		*xstrdup(const char *);

void		 dump_relocations(Obj_Entry *);
void		 dump_obj_relocations(Obj_Entry *);
void		 dump_Elf_Rel(Obj_Entry *, const Elf_Rel *, u_long);
void		 dump_Elf_Rela(Obj_Entry *, const Elf_Rela *, u_long);

/*
 * Function declarations.
 */
const char	*basename(const char *);
int		 do_copy_relocations(Obj_Entry *);
unsigned long	 elf_hash(const char *);
const Elf_Sym	 *find_symdef(unsigned long, const Obj_Entry *,
			      const Obj_Entry **, bool, SymCache *);
void		 init_pltgot(Obj_Entry *);
void		 lockdflt_init(LockInfo *);
void		 Tcl_ObjFree(Obj_Entry *);
Obj_Entry	*Tcl_ObjNew(void);
int		 reloc_non_plt(Obj_Entry *, Obj_Entry *);
int		 reloc_plt(Obj_Entry *);
int		 reloc_jmpslots(Obj_Entry *);
void		 _rtld_bind_start(void);
const Elf_Sym	*Tcl_SymlookObj(const char *, unsigned long, const Obj_Entry *,
			     bool);

void		*Tcl_TlsGetAddrCommon(void **, int, size_t);
struct tls_tcb	*allocate_tls(Obj_Entry *);
void		 free_tls(struct tls_tcb *);
void		 *Tcl_AllocateModuleTls(int);
bool		 Tcl_AllocateTlsOffset(Obj_Entry *);
void		 Tcl_FreeTlsOffset(Obj_Entry *);
void		 allocate_initial_tls(Obj_Entry *);

void	*xcalloc(size_t);
void	*xmalloc(size_t);
void	*xrealloc(void *, size_t);
char	*xstrdup(const char *);

void *Tcl_Dlopen(const void *addr, int nbytes, int mode);
void *Tcl_Dlsym(void *handle, const char *name);
int Tcl_Dlclose(void *handle);
void Tcl_InitRtld(void);

#endif

